use std::collections::HashMap;
use std::fmt;

use crate::utils::rss;
use crate::{datasources::posts::PostsDatasource, state::AppState};
use axum::http::{HeaderMap, HeaderValue};
use axum::{
    extract::{Path, State},
    http::{header, StatusCode},
    response::IntoResponse,
    routing::get,
    Json, Router,
};
use chrono::Utc;
use fred::types::Expiration;
use serde::{Deserialize, Deserializer, Serialize, Serializer};

#[derive(sqlx::FromRow, Deserialize, Serialize, Debug, Clone)]
pub struct Post {
    pub post_id: i32,
    pub author_id: Option<i32>,
    pub first_name: Option<String>,
    pub last_name: Option<String>,
    pub title: String,
    pub body: String,
    #[serde(serialize_with = "serialize_datetime")]
    #[serde(deserialize_with = "deserialize_datetime")]
    pub created_at: Option<chrono::DateTime<Utc>>,
}

#[derive(sqlx::FromRow, Deserialize, Serialize, Debug, Clone)]
pub struct PostFeaturedVariant {
    pub post_id: i32,
    pub author_id: Option<i32>,
    pub first_name: Option<String>,
    pub last_name: Option<String>,
    pub title: String,
    pub body: String,
    #[serde(serialize_with = "serialize_datetime")]
    #[serde(deserialize_with = "deserialize_datetime")]
    pub created_at: Option<chrono::DateTime<Utc>>,
    pub is_featured: Option<bool>,
}

#[derive(Deserialize)]
pub struct PostGetOneParams {
    pub id: i32,
}

pub struct PostsRoute;
impl PostsRoute {
    pub fn routes(app_state: &AppState) -> Router {
        // add more post routes here!
        Router::new()
            .route("/all", get(PostsRoute::get_all))
            .route("/:id", get(PostsRoute::get_one))
            .route("/recent", get(PostsRoute::get_recent_posts))
            .route("/popular", get(PostsRoute::get_popular_posts))
            .route("/hot", get(PostsRoute::get_hot_posts))
            .route("/featured", get(PostsRoute::get_featured_posts))
            .route("/rss", get(PostsRoute::get_rss_posts))
            .with_state(app_state.clone())
    }

    // get all posts
    async fn get_all(State(app_state): State<AppState>) -> impl IntoResponse {
        let mut state = app_state.lock().await;
        let cached: Option<Vec<Post>> = state
            .cache
            .get(String::from("posts:all"))
            .await
            .unwrap_or(None);

        if let Some(posts) = cached {
            tracing::info!("grabbing all posts from cache");
            return Ok(Json(posts));
        };

        match PostsDatasource::get_all(&state.database).await {
            Ok(posts) => {
                tracing::info!("grabbing all posts from database");
                if let p = &posts {
                    let posts = p.clone();
                    let state = app_state.clone();

                    tracing::info!("storing database data in cache");
                    tokio::spawn(async move {
                        let mut s = state.lock().await;
                        let _ = s
                            .cache
                            .set(
                                String::from("posts:all"),
                                &posts,
                                Some(Expiration::EX(10)),
                                None,
                                false,
                            )
                            .await;
                    });
                };

                Ok(Json(posts))
            }
            Err(e) => Err((StatusCode::INTERNAL_SERVER_ERROR, e.to_string())),
        }
    }

    // get one post
    async fn get_one(
        State(app_state): State<AppState>,
        Path(params): Path<PostGetOneParams>,
    ) -> impl IntoResponse {
        let mut state = app_state.lock().await;
        let cached: Option<PostFeaturedVariant> = state
            .cache
            .get(format!("posts:{}", params.id))
            .await
            .unwrap_or(None);

        if let Some(post) = cached {
            tracing::info!("grabbing one post from cache");
            return Ok(Json(post));
        };

        match PostsDatasource::get_one(&state.database, params.id).await {
            Ok(post) => {
                tracing::info!("grabbing one post from database");
                if let p = &post {
                    let post = p.clone();
                    let state = app_state.clone();

                    tracing::info!("storing database data in cache");
                    tokio::spawn(async move {
                        let mut s = state.lock().await;
                        let _ = s
                            .cache
                            .set(
                                format!("posts:{}", params.id),
                                &post,
                                Some(Expiration::EX(10)),
                                None,
                                false,
                            )
                            .await;
                    });
                };

                Ok(Json(post))
            }
            Err(e) => Err((StatusCode::INTERNAL_SERVER_ERROR, e.to_string())),
        }
    }

    // get recent posts
    async fn get_recent_posts(State(app_state): State<AppState>) -> impl IntoResponse {
        let mut state = app_state.lock().await;
        let cached: Option<Vec<Post>> = state
            .cache
            .get(String::from("posts:recent"))
            .await
            .unwrap_or(None);

        if let Some(posts) = cached {
            tracing::info!("grabbing recent posts from cache");
            return Ok(Json(posts));
        };

        match PostsDatasource::get_recent(&state.database).await {
            Ok(posts) => {
                tracing::info!("grabbing recent posts from database");
                if let p = &posts {
                    let posts = p.clone();
                    let state = app_state.clone();

                    tracing::info!("storing database data in cache");
                    tokio::spawn(async move {
                        let mut s = state.lock().await;
                        let _ = s
                            .cache
                            .set(
                                String::from("posts:recent"),
                                &posts,
                                Some(Expiration::EX(5)),
                                None,
                                false,
                            )
                            .await;
                    });
                };

                Ok(Json(posts))
            }
            Err(e) => Err((StatusCode::INTERNAL_SERVER_ERROR, e.to_string())),
        }
    }

    // get the top three posts with the highest view count
    async fn get_popular_posts(State(app_state): State<AppState>) -> impl IntoResponse {
        let mut state = app_state.lock().await;
        let cached: Option<Vec<Post>> = state
            .cache
            .get(String::from("posts:popular"))
            .await
            .unwrap_or(None);

        if let Some(posts) = cached {
            tracing::info!("grabbing popular posts from cache");
            return Ok(Json(posts));
        };

        match PostsDatasource::get_popular(&state.database).await {
            Ok(posts) => {
                tracing::info!("grabbing popular posts from database");
                if let p = &posts {
                    let posts = p.clone();
                    let state = app_state.clone();

                    tracing::info!("storing database data in cache");
                    tokio::spawn(async move {
                        let mut s = state.lock().await;
                        let _ = s
                            .cache
                            .set(
                                String::from("posts:popular"),
                                &posts,
                                Some(Expiration::EX(5)),
                                None,
                                false,
                            )
                            .await;
                    });
                };

                Ok(Json(posts))
            }
            Err(e) => Err((StatusCode::INTERNAL_SERVER_ERROR, e.to_string())),
        }
    }

    // get the top three posts with the most comments
    async fn get_hot_posts(State(app_state): State<AppState>) -> impl IntoResponse {
        let mut state = app_state.lock().await;
        let cached: Option<Vec<Post>> = state
            .cache
            .get(String::from("posts:hot"))
            .await
            .unwrap_or(None);

        if let Some(posts) = cached {
            tracing::info!("grabbing hot posts from cache");
            return Ok(Json(posts));
        };

        match PostsDatasource::get_hot(&state.database).await {
            Ok(posts) => {
                tracing::info!("grabbing hot posts from database");
                if let p = &posts {
                    let posts = p.clone();
                    let state = app_state.clone();

                    tracing::info!("storing database data in cache");
                    tokio::spawn(async move {
                        let mut s = state.lock().await;
                        let _ = s
                            .cache
                            .set(
                                String::from("posts:hot"),
                                &posts,
                                Some(Expiration::EX(5)),
                                None,
                                false,
                            )
                            .await;
                    });
                };

                Ok(Json(posts))
            }
            Err(e) => Err((StatusCode::INTERNAL_SERVER_ERROR, e.to_string())),
        }
    }

    // get posts that are featured
    async fn get_featured_posts(State(app_state): State<AppState>) -> impl IntoResponse {
        let mut state = app_state.lock().await;
        let cached: Option<Vec<Post>> = state
            .cache
            .get(String::from("posts:featured"))
            .await
            .unwrap_or(None);

        if let Some(posts) = cached {
            tracing::info!("grabbing featured posts from cache");
            return Ok(Json(posts));
        };

        match PostsDatasource::get_featured(&state.database).await {
            Ok(posts) => {
                tracing::info!("grabbing featured posts from database");
                if let p = &posts {
                    let posts = p.clone();
                    let state = app_state.clone();

                    tracing::info!("storing database data in cache");
                    tokio::spawn(async move {
                        let mut s = state.lock().await;
                        let _ = s
                            .cache
                            .set(
                                String::from("posts:featured"),
                                &posts,
                                Some(Expiration::EX(5)),
                                None,
                                false,
                            )
                            .await;
                    });
                };

                Ok(Json(posts))
            }
            Err(e) => Err((StatusCode::INTERNAL_SERVER_ERROR, e.to_string())),
        }
    }

    // get rss posts
    async fn get_rss_posts(State(app_state): State<AppState>) -> impl IntoResponse {
        let state = app_state.lock().await;

        match PostsDatasource::get_all(&state.database).await {
            Ok(posts) => {
                let web_url = std::env::var("BASE_URI_WEB").expect("No environment variable found");
                let mapped_posts: HashMap<String, Post> = posts
                    .into_iter()
                    .map(|post| (post.post_id.to_string(), post))
                    .collect();
                let xml: String = rss::generate_rss(
                    "Wyatt's blog",
                    "Wyatt and friends",
                    web_url.as_str(),
                    &mapped_posts,
                );
                let mut headers = HeaderMap::new();
                headers.insert(
                    header::CONTENT_DISPOSITION,
                    HeaderValue::from_str(r#"attachment; filename="posts.xml""#).unwrap(),
                );
                headers.insert(
                    header::CONTENT_TYPE,
                    HeaderValue::from_static("application/xml"),
                );
                (headers, xml)
            }
            Err(e) => {
                let mut headers = HeaderMap::new();
                headers.insert("Content-Type", HeaderValue::from_static("text/plain"));
                (headers, e.to_string())
            }
        }
    }
}

pub fn serialize_datetime<S>(
    date: &Option<chrono::DateTime<Utc>>,
    serializer: S,
) -> Result<S::Ok, S::Error>
where
    S: Serializer,
{
    serializer.serialize_str(&date.unwrap().to_rfc3339())
}

pub fn deserialize_datetime<'de, D>(
    deserializer: D,
) -> Result<Option<chrono::DateTime<Utc>>, D::Error>
where
    D: Deserializer<'de>,
{
    struct DateTimeVisitor;

    impl<'de> serde::de::Visitor<'de> for DateTimeVisitor {
        type Value = Option<chrono::DateTime<Utc>>;

        fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
            formatter.write_str("an ISO 8601 formatted date string or null")
        }

        fn visit_str<E>(self, value: &str) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            match chrono::DateTime::parse_from_rfc3339(value) {
                Ok(dt) => Ok(Some(dt.with_timezone(&Utc))),
                Err(e) => Err(E::custom(format!("Failed to parse datetime: {}", e))),
            }
        }

        fn visit_none<E>(self) -> Result<Self::Value, E>
        where
            E: serde::de::Error,
        {
            Ok(None)
        }

        fn visit_some<D>(self, deserializer: D) -> Result<Self::Value, D::Error>
        where
            D: Deserializer<'de>,
        {
            deserializer.deserialize_str(self)
        }
    }

    deserializer.deserialize_option(DateTimeVisitor)
}
